#include "watchdog.h"

#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <dirent.h>
#include <linux/limits.h>
#include <unistd.h>

#define timespec_sec(n) ((struct timespec){.tv_sec = (n)})

#define foreach_dir(ent, dp)                                                   \
  for (struct dirent *ent = readdir(dp); ent != NULL; ent = readdir(dp))

bool find_dropbear(void);
char *get_proc_name(const char *);
void watch_init(void);
void _dummy(int);

void dropbear_watch(void) {
  watch_init();

  for (;;) {
    bool found = find_dropbear();
    if (!found)
      system("/usr/sbin/dropbear");

    printf("dropbear found: %s\n", found ? "true" : "false");

    nanosleep(&timespec_sec(found ? 60 : 15), NULL);
  }
}

void self_watch(pid_t dropbear_watch_pid) {
  watch_init();

  for (;;) {
    nanosleep(&timespec_sec(60), NULL);

    printf("bark\n");

    DIR *dp = opendir("/proc");
    if (dp == NULL)
      continue;

    bool found = false;

    foreach_dir(ent, dp) {
      long lpid = atol(ent->d_name);
      if (lpid <= 0)
        continue;

      if (lpid == dropbear_watch_pid)
        found = true;

      if (found)
        break;
    }

    if (!found) {
      dropbear_watch_pid = fork();
      if (dropbear_watch_pid != 0)
        dropbear_watch();
    }
  }
}

bool find_dropbear() {
  const char DROPBEAR_NAME[] = "dropbear";

  DIR *dp = opendir("/proc");
  if (dp == NULL)
    return false;

  bool found = false;

  foreach_dir(ent, dp) {
    long lpid = atol(ent->d_name);
    if (lpid <= 0)
      continue;

    char *name = get_proc_name(ent->d_name);
    if (name == NULL)
      continue;

    char *found_dropbear = strstr(name, DROPBEAR_NAME);
    if (found_dropbear != NULL)
      found = true;

    free(name);

    if (found)
      break;
  }

  closedir(dp);

  return found;
}

char *get_proc_name(const char *pid) {
  const size_t LINE_MAX = 1024;

  char path[PATH_MAX];
  snprintf(path, PATH_MAX, "/proc/%s/status", pid);

  FILE *fp = fopen(path, "r");
  if (fp == NULL)
    return NULL;

  char s[LINE_MAX], key[LINE_MAX], val[LINE_MAX];
  while (fscanf(fp, "%s %s", key, val) != EOF) {
    if (strcmp(key, "Name:") == 0)
      break;
  }

  char *name = strdup(val);

  return name;
}

void watch_init(void) {
  setsid();

  struct sigaction act = {.sa_flags = 0, .sa_handler = _dummy};
  int sigs[] = {SIGINT, SIGTERM, SIGQUIT, SIGSTOP, SIGCHLD};
  for (size_t i = 0; i < sizeof sigs / sizeof sigs[0]; i++)
    sigaction(sigs[i], &act, NULL);
}

void _dummy(int arg) {
  (void)arg;

  // do nothing
}
